/* -*- Mode:C++; -*- */

// coop.cc

#include <iostream>
#include <NTL/mat_ZZ.h>

#include "QQ.h"
#include "rig_maxdet.h"
#include "covering.h"
#include "interior_point.h"

#define COUT_PRECISION 10

// The setting of this parameter can cause trouble!
#define BEYOND 0.00001    

void die_coop(const char* last_words)
{
  std::cout <<"coop.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

// ==========
// === co ===
// === op ===
// ==========

int main(int argc, char** argv)
{
  int output_level = OUTPUT_LEVEL_COMMUNICATIVE;
  bool read_and_exclude_boundary_facets = false;
  bool certify_local_optimality = false;
  bool set_duality_gap = false; 
  bool read_interior_point = false;
  bool set_maximum_iterations = false;
  bool quick_and_dirty = false;

  int max_iterations; // = 100;
  double duality_gap; // = 0.000000001; 

  int c;
  while ((c = getopt (argc, argv, "hbcd:im:qvn")) != -1)
    switch (c)
      {
      case 'h':
	std::cout <<"Usage: coop [options] <file\n";
	std::cout <<"Options:\n";
	std::cout <<" -h   show this help\n";
	std::cout <<" -b   read and exclude boundary facets\n";
	std::cout <<" -c   certify local optimality\n";
	std::cout <<" -d x set duality gap to x\n";
	std::cout <<" -i   read interior point\n";
        std::cout <<" -m N set maximum iterations to N\n";
	std::cout <<" -q   quick and dirty (non-rigorous)\n";
	std::cout <<" -v   be verbose\n";
	std::cout <<" -n   be non-verbose\n";
	return 0;
	break;
      case 'b':
	read_and_exclude_boundary_facets = true;
	break;
      case 'c':
	certify_local_optimality = true;
	break;
      case 'd':
	duality_gap = atof(optarg);
	set_duality_gap = true; 	
	break;
      case 'i':
	read_interior_point = true;
	break;
      case 'm':
	max_iterations = atoi(optarg);
	set_maximum_iterations = true; 
	break;
      case 'q':
	quick_and_dirty = true; 
	break;
      case 'v':
	output_level = OUTPUT_LEVEL_ALL;
	break;
      case 'n':
	output_level = OUTPUT_LEVEL_UNCOMMUNICATIVE;
	break;
      default:
	std::cout <<"coop: invalid option\n";
	std::cout <<"Try 'coop -h' for more information.\n";
	return 1;
      }
  
  if (quick_and_dirty && certify_local_optimality) {
    std::cout <<"coop: option -c has no effect if envoked with option -q\n";
    std::cout <<"Try 'coop -h' for more information.\n";
    return 1;
  }

  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"*** co ***\n";
    std::cout <<"*** op ***\n\n";
  }

  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"* READING INPUT, expecting:\n";
    std::cout <<"- dim (int)\n";
    std::cout <<"- simplex_count (int)\n";
    std::cout <<"- simplex (int[simplex_count*dim*dim])\n";
    std::cout <<"- basis_count (int)\n";
    std::cout <<"- basis (int[basis_count*dim*(dim+1)/2])\n";
    std::cout <<"- inequality_count (int)\n";
    if (read_and_exclude_boundary_facets)
      std::cout <<"- boundary_facets (bool[inequality_count])\n";
    std::cout <<"- inequality (int[inequality_count*dim])\n";
    if (read_interior_point)
      std::cout <<"- interior (double[dim])\n";
    if (!set_maximum_iterations)
      std::cout <<"- max_iterations (int)\n";
    if (!set_duality_gap)
      std::cout <<"- duality_gap (double)\n";
    std::cout << std::endl;
  }

  // 1. read input

  int dim;
  std::cin >>dim;

  int simplex_count;
  std::cin >>simplex_count;
  int* simplex;
  simplex = (int*) calloc(simplex_count*dim*dim, sizeof(int));
  if (simplex == NULL) die_coop("simplex == NULL.\n");
  for (int i = 0; i < simplex_count*dim*dim; i++)
    std::cin >>simplex[i];

  int basis_count;
  std::cin >>basis_count;
  int* basis;
  basis = (int*) calloc(basis_count*dim*(dim+1)/2, sizeof(int));
  if (basis == NULL) die_coop("basis == NULL.\n");
  for (int i = 0; i < basis_count*dim*(dim+1)/2; i++) 
    std::cin >>basis[i];

  int inequality_count;
  std::cin >>inequality_count;

  bool* boundary_facets;
  boundary_facets = (bool*) calloc(inequality_count, sizeof(bool));
  if (boundary_facets == NULL) die_coop("boundary_facets == NULL.\n");
  if (read_and_exclude_boundary_facets) {
    for (int i = 0; i < inequality_count; i++) 
      std::cin >>boundary_facets[i];
  }
  else {
    for (int i = 0; i < inequality_count; i++) 
      boundary_facets[i] = false;
  }

  int* inequality;
  inequality = (int*) calloc(inequality_count*basis_count, sizeof(int));
  if (inequality == NULL) die_coop("inequality == NULL.\n");
  for (int i = 0; i < inequality_count*basis_count; i++) 
    std::cin >>inequality[i];

  maxdet_data md;
  setup_maxdet_covering_constraints(dim, simplex_count, simplex, basis_count, basis, inequality_count, inequality, md);  
  setup_maxdet_covering_cost_func(dim, basis_count, basis, md);
  
  double* interior;
  interior = (double*) calloc(basis_count, sizeof(double));
  if (interior == NULL) die_coop ("interior == NULL.\n");
  if (read_interior_point) {    
    for (int i = 0; i < basis_count; i++) 
      std::cin >>interior[i];
  }

  if (!set_maximum_iterations)
    std::cin >>max_iterations;

  if (!set_duality_gap)
    std::cin >>duality_gap;

  // interior point is computed after input has been read, since lrs may manipulate std::cin
  if (!read_interior_point) {
    if (!compute_interior_point(basis_count,inequality_count,inequality,interior))
      die_coop("Could not obtain interior point.\n");
    // normalize maximum diagonal entry to 1 (e.g. to ensure mu(interior)<=1)    
    double max_diag = 0.0;
    for (int i = 0; i < dim; i++) {
      double i_diag_entry = 0.0;
      for (int basis_index = 0; basis_index < basis_count; basis_index++) {
	i_diag_entry += interior[basis_index] * double(basis[basis_index*dim*(dim+1)/2+POSLT(dim,i,i)]);
      }
      if (i_diag_entry > max_diag)
        max_diag = i_diag_entry;
    }
    if (max_diag == 0.0)
      die_coop("Obtaining interior point failed.\n");
    for (int i = 0; i < basis_count; i++)
      interior[i] = interior[i] / max_diag;  

  }

  std::cout.precision(COUT_PRECISION);
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    if (!read_interior_point) {
      std::cout <<"\n* computed interior point\n";
      for (int i = 0; i < basis_count; i++) 
	std::cout <<interior[i] <<" ";
      std::cout <<std::endl;
    }
    std::cout <<std::endl;
  }

  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"\n";
    std::cout <<"* dimension = " <<dim <<"\n";
    std::cout <<"* simplex_count = " <<simplex_count <<"\n";
    std::cout <<"* basis_count = " <<basis_count <<"\n";
    std::cout <<"* inequality_count = " <<inequality_count <<"\n";
    std::cout <<"* max_iterations = " <<max_iterations <<"\n";
    std::cout <<"* duality_gap = " <<duality_gap <<"\n";
    std::cout <<"\n";
  }

  // 2. solve MAXDET quick and dirty or rigourously

  if (quick_and_dirty) {
    double lower_bound;
    double upper_bound;

    solve_maxdet_nonrigorously(md, interior, duality_gap, max_iterations, lower_bound, upper_bound, output_level-1);
   
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      std::cout <<"\n* theta_lower_bound ~ " << 1/sqrt(exp(-lower_bound)) <<"\n";
      std::cout <<"* theta_upper_bound ~ " << 1/sqrt(exp(-upper_bound)) <<"\n";
    }

    free_maxdet_covering_data(md);
    free(boundary_facets);
    free(interior);
    free(inequality);
    free(basis);
    free(simplex);
    return 0;
  }

  // Continue rigourously
  NTL::vec_ZZ minimizer_approx;
  NTL::vec_ZZ dual_Z_approx;       
  NTL::vec_ZZ dual_W_approx;       
  maxdet_bound lower_bound;
  maxdet_bound upper_bound;

  solve_maxdet_rigorously(md, interior, duality_gap, max_iterations, minimizer_approx, dual_Z_approx, dual_W_approx, lower_bound, upper_bound, output_level);

  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    QQ lin_part_diff;
    QQ det_part_diff;
    lin_part_diff = upper_bound.lin_part - lower_bound.lin_part;
    det_part_diff = upper_bound.det_part / lower_bound.det_part;
	
    if ((lin_part_diff == NTL::to_ZZ(0)) && (det_part_diff == NTL::to_ZZ(1))) {
      if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
	if ((lin_part_diff == 0) && (det_part_diff == 1))  
	  std::cout << "* CERTIFIED OPTIMUM *\n";
	
	if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	  std::cout << "------ Proof -------\n"; 
	  std::cout <<md.m <<"\n\n";
	  
	  std::cout <<md.L <<"\n";
	  for (int i=0; i<md.L; i++)  
	    std::cout << md.F_blkszs[i] << " ";
	  std::cout << std::endl;
	  std::cout <<md.F << "\n\n";
	  
	  std::cout <<md.K <<"\n";
	  for (int i=0; i<md.K; i++)  
	    std::cout << md.G_blkszs[i] << " ";
	  std::cout << std::endl;
	  std::cout <<md.G << "\n\n";

	  std::cout <<md.c << "\n\n";
	  
	  std::cout <<minimizer_approx << "\n";
	  std::cout <<dual_Z_approx << "\n";
	  std::cout <<dual_W_approx << "\n";
	  std::cout << "------- qed --------\n"; 
	}
	else
	  std::cout << "* NOT A CERTIFIED OPTIMUM *\n\n";
      }
    }
  }

  // 3. obtain additional certificates

  bool lies_on_boundary = false;
  bool probably_on_facets = false;
  if (certify_local_optimality) {

    if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
      std::cout <<"* TRYING TO OBTAIN ADDITIONAL CERTIFICATES\n";
    }

    // 3.A. without linear constraints: (sec_upper_bound <= lower_bound)  implies "opt in bd(cone)"
    maxdet_data sec_md;
    setup_maxdet_covering_constraints(dim, simplex_count, simplex, basis_count, basis, 0, NULL, sec_md);
    setup_maxdet_covering_cost_func(dim, basis_count, basis, sec_md);
    NTL::vec_ZZ sec_minimizer_approx;
    NTL::vec_ZZ sec_dual_Z_approx;       
    NTL::vec_ZZ sec_dual_W_approx;       
    maxdet_bound sec_lower_bound;
    maxdet_bound sec_upper_bound;
    solve_maxdet_rigorously(sec_md, interior, duality_gap, max_iterations, sec_minimizer_approx,
			    sec_dual_Z_approx, sec_dual_W_approx, sec_lower_bound, sec_upper_bound, output_level-1);
    free_maxdet_covering_data(sec_md);

    if (lower_bound > sec_upper_bound) {
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	std::cout <<"* CERTIFIED \"opt lies on boundary\"\n";
      }
      lies_on_boundary = true;
    }
    else { 
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	std::cout <<"* NO CERTIFICATE for \"opt lies on boundary\"\n";
      }
    }
    
    // 3.B. for each linear constraint (facet of cone), 
    //      determine minimum "on the other side", if it is not a boundary facet
    //      (sec_lower_bound > upper_bound) implies "opt not on facet"
    for (int ineq_index = 0; ineq_index < inequality_count; ineq_index++) 
      if (!boundary_facets[ineq_index]) {
	// B.I. construct sec_interior as interior point of facet and "beyond a little" 
	double* sec_interior;
	sec_interior = (double*) calloc(basis_count, sizeof(double));
	if (sec_interior == NULL) die_coop ("sec_interior == NULL.\n");

	if (!compute_relative_interior_point_in_facet(basis_count,inequality_count,inequality,ineq_index,sec_interior))
	  die_coop("Could not obtain interior point.\n");

	double in_pr_ineq_ineq = 0.0;
	for (int i = 0; i < basis_count; i++) 
	  in_pr_ineq_ineq  += double(inequality[ineq_index*basis_count + i])*double(inequality[ineq_index*basis_count + i]);
	
	for (int i = 0; i < basis_count; i++) 
	  sec_interior[i] = sec_interior[i] - BEYOND / sqrt(in_pr_ineq_ineq) * double(inequality[ineq_index*basis_count + i]);

	// normalize maximum diagonal entry to 1
	double max_diag = 0.0;
	for (int i = 0; i < dim; i++) {
	  double i_diag_entry = 0.0;
	  for (int basis_index = 0; basis_index < basis_count; basis_index++) {
	    i_diag_entry += sec_interior[basis_index] * double(basis[basis_index*dim*(dim+1)/2+POSLT(dim,i,i)]);
	  }
	  if (i_diag_entry > max_diag)
	    max_diag = i_diag_entry;
	}
	if (max_diag == 0.0)
	  die_coop("Obtaining interior point failed.\n");
	for (int i = 0; i < basis_count; i++)
	  sec_interior[i] = sec_interior[i] / max_diag;  
	
	// B.II. setup MAXDET with one inequality (switched)  ineq_index
	int* sec_inequality = (int*) calloc(basis_count, sizeof(int));
	if (inequality == NULL) die_coop("sec_inequality == NULL.\n");
	for (int basis_index = 0; basis_index < basis_count; basis_index++) 
	  sec_inequality[basis_index] = -inequality[ineq_index*basis_count + basis_index];
	
	setup_maxdet_covering_constraints(dim, simplex_count, simplex, basis_count, basis, 1, sec_inequality, sec_md);
	setup_maxdet_covering_cost_func(dim, basis_count, basis, sec_md);
	solve_maxdet_rigorously(sec_md, sec_interior, duality_gap, max_iterations, sec_minimizer_approx,
				sec_dual_Z_approx, sec_dual_W_approx, sec_lower_bound, sec_upper_bound, output_level-1);
	free_maxdet_covering_data(sec_md);
	free(sec_inequality);
	free(sec_interior);
	
	if (sec_lower_bound > upper_bound) {
	  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	    std::cout <<"* CERTIFIED \"opt does not lie on facet " <<ineq_index+1 <<"\"\n";
	  }
	}
	else {
	  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) { 
	    std::cout <<"* NO CERTIFICATE for \"opt not on facet " <<ineq_index+1 <<"\"\n";
	  }
	  probably_on_facets = true;
	}
      }
  }

  // 4. summary 

  QQ lin_part_diff;
  QQ det_part_diff;
  lin_part_diff = upper_bound.lin_part - lower_bound.lin_part;
  det_part_diff = upper_bound.det_part / lower_bound.det_part;

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"\n";
    std::cout <<"***********\n";
    std::cout <<"* SUMMARY *\n";
    std::cout <<"***********\n";
    std::cout <<"\n";
  }
  
  if (certify_local_optimality) {
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      if (lies_on_boundary && !probably_on_facets)
	std::cout << "* CERTIFICATES CONTRADICT EACH OTHER (indicating approximation problems)\n";
      if (lies_on_boundary)
	std::cout << "* CERTIFICATE for optimum in bd(cone)\n";
      if (!probably_on_facets) {
	std::cout << "* CERTIFICATE for optimum in int(cone)\n";
        if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE && read_and_exclude_boundary_facets) {
          std::cout << "  -- assuming boundary facets in input contain non-positive forms only --\n";
	}
      }
      if (probably_on_facets && !lies_on_boundary) {
	std::cout << "* NO CERTIFICATE for optimum in bd(cone) or optimum in int(cone)\n";
        std::cout << "  -- optimum might be exactly on boundary --\n";
      }
    }
  }

  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"* minimizer_approx = " <<minimizer_approx <<"\n";
  }

  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"* dual_Z_approx = " <<dual_Z_approx <<"\n";
    std::cout <<"* dual_W_approx = " <<dual_W_approx <<"\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"* theta_lower_bound";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<" = 1/sqrt(exp(" << -lower_bound.lin_part << " - log("<< lower_bound.det_part << ")))\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<" ~ " << 1/sqrt(exp(-to_double(lower_bound.lin_part) - log(to_double(lower_bound.det_part)))) << "\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"* theta_upper_bound";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<" = 1/sqrt(exp(" << -upper_bound.lin_part << " - log("<< upper_bound.det_part << ")))\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<" ~ " << 1/sqrt(exp(-to_double(upper_bound.lin_part) - log(to_double(upper_bound.det_part)))) << "\n";
  }

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"* duality_gap";
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<" = " <<lin_part_diff <<" + log(" <<det_part_diff <<")\n";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {  
    std::cout <<" ~ " <<to_double(lin_part_diff) + log(to_double(det_part_diff))  <<"\n" ;
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"* Thanks for your patience...\n";
  }

  // 5. clean up

  free_maxdet_covering_data(md);  
  free(boundary_facets);
  free(interior);
  free(inequality);
  free(basis);
  free(simplex);

  return 0;
}

