/* -*- Mode:C++; -*- */

// covering.cc

#include <NTL/mat_ZZ.h>
#include "covering.h"
#include "rig_maxdet.h"

void die_covering(const char* last_words)
{
  std::cout <<"covering.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

int compute_F_entry
(
 const int dim,
 const int* basis,
 const int basis_index,
 const int* simplex,
 const int simplex_index,
 const int row,
 const int col
)
{
  if (row < col) die_covering("row < col.\n");
  if (row == 0) return 0;

  int k, l;

  if (col == 0) { 
    k = row - 1; 
    l = row - 1; 
  } 
  else {
    k = row - 1;
    l = col - 1;
  }

  // entry = sum_{1 \leq i, j, \leq dim} v_i^t B v_j
  // v_i, v_j - vectors of simplex[simplex_index*dim*dim)]
  // B - basis vector basis[basis_index*((dim*(dim+1))/2] 

  int entry = 0;
  int bij;
  for (int i = 0; i < dim; i++)
    for (int j = 0; j < dim; j++) { 
      if (j <= i)
	bij = basis[basis_index*(dim*(dim+1))/2+POSLT(dim,i,j)];
      else 
	bij = basis[basis_index*(dim*(dim+1))/2+POSLT(dim,j,i)];
      entry = entry +
	simplex[simplex_index*dim*dim+k*dim+i] *
	bij *
	simplex[simplex_index*dim*dim+l*dim+j];	
    }
  return entry;
}

void setup_maxdet_covering_constraints 
(
 const int dim,
 const int simplex_count,
 const int* simplex,
 const int basis_count,
 const int* basis,
 const int inequality_count,
 const int* inequality,
 maxdet_data& md
)
{
  md.m = basis_count;
  md.L = simplex_count+inequality_count;

  md.F_blkszs = (int*) calloc(md.L, sizeof(int));
  if (md.F_blkszs == NULL) die_covering("md.F_blkszs == NULL.\n");
  for (int i = 0; i < simplex_count; i++) md.F_blkszs[i] = dim + 1;
  for (int i = 0; i < inequality_count; i++) md.F_blkszs[simplex_count+i] = 1;

  int F_sz = simplex_count*((dim+1)*(dim+2)/2)+inequality_count;
  md.F.SetLength((basis_count+1)*F_sz + 1);
  md.F[0] = 1;
  // F - semidefinite conditions
  for (int simplex_index = 0; simplex_index < simplex_count; simplex_index++)
    md.F[1 + simplex_index*((dim+1)*(dim+2)/2)] = 4;
  for (int basis_index = 0; basis_index < basis_count; basis_index++) 
    for (int simplex_index = 0; simplex_index < simplex_count; simplex_index++) 
      for (int row = 0; row < dim+1; row++)
	for (int col = 0; col <= row; col++) 
	  md.F[1 + (basis_index+1)*F_sz + simplex_index*((dim+1)*(dim+2)/2) + 
	       POSUT(dim+1,col,row)] = 
	    NTL::to_ZZ(compute_F_entry(dim, basis, basis_index, simplex, simplex_index,
				  row, col));
  // F - linear conditions
  for (int inequality_index = 0; inequality_index < inequality_count; inequality_index++)
    for (int basis_index = 0; basis_index < basis_count; basis_index++) 
      md.F[1 + (basis_index+1)*F_sz + simplex_count*((dim+1)*(dim+2)/2) + inequality_index] =
	NTL::to_ZZ(inequality[inequality_index*(basis_count) + basis_index]);
}

void setup_maxdet_covering_cost_func 
(
 const int dim,
 const int basis_count,
 const int* basis,
 maxdet_data& md
) 
{
  md.K = 1;
  md.G_blkszs = (int*) calloc(md.K, sizeof(int));
  if (md.G_blkszs == NULL) die_covering("md.G_blkszs == NULL.\n");
  md.G_blkszs[0] = dim;

  int G_sz = dim*(dim+1)/2;
  md.G.SetLength((basis_count+1)*G_sz + 1);
  md.G[0] = 1;
  for (int basis_index = 0; basis_index < basis_count; basis_index++)
    for (int row = 0; row < dim; row++)
      for (int col = 0; col <= row; col++)
	md.G[1 + (basis_index+1)*G_sz + POSUT(dim,col,row)] = 
	  NTL::to_ZZ(basis[basis_index*(dim*(dim+1))/2 + POSLT(dim,row,col)]);

  md.c.SetLength(basis_count + 1);
  md.c[0] = 1;
}

void setup_maxdet_packing_covering_cost_func 
(
 const int dim,
 const int basis_count,
 maxdet_data& md
) 
{
  md.K = 1;
  md.G_blkszs = (int*) calloc(md.K, sizeof(int));
  if (md.G_blkszs == NULL) die_covering("md.G_blkszs == NULL.\n");
  md.G_blkszs[0] = 1;
  md.G.SetLength((basis_count+1) + 1);
  md.G[0] = 1; // set common denominator to 1
  md.G[1] = 1; // so constantly det G = 1
  for (int basis_index = 0; basis_index < basis_count; basis_index++)
    md.G[2 + basis_index] = 0;
  md.c.SetLength(basis_count + 1); 
  md.c[0] = 1;
  md.c[basis_count] = -1;
}

void free_maxdet_covering_data
(
 maxdet_data& md
)
{
  free(md.F_blkszs);
  free(md.G_blkszs);
}
