/* -*- Mode:C++; -*- */

// interior_point.cc

#include <iostream>
#include <cstdio>
#include <cstring>

#include "interior_point.h"
#include "gmp.h"
#include "lrslib.h"

void die_interior_point(const char* last_words)
{
  std::cerr <<"interior_point.cc: "<<last_words;
  exit(EXIT_FAILURE);
}

//
// compute_interior_point 
//
// -- returns true and writes solution of system 
//    sum_j ineq[i][j] x_j > 0 into interior if one exists, 
// -- returns false otherwise
// -- exits with an error message if system conatins linearity space. 
//

bool compute_interior_point 
(
 const int dim,
 const int inequality_count,
 const int* inequality,
 double* interior
 )
{
  if (!lrs_init ("interior_point")) 
    die_interior_point("could not initialize lrs.\n");
 
  lrs_dic *P;
  lrs_dat *Q;
       
  Q = lrs_alloc_dat("LRS globals");
  if (Q == NULL) die_interior_point("Q == NULL.\n");
  Q->m = inequality_count;
  Q->n = dim+1;
  Q->verbose = FALSE;
  //Q->debug=TRUE;   
    
  P = lrs_alloc_dic(Q);
  if (P == NULL) die_interior_point("p == NULL.\n");
    
  long den[dim+1];
  long enu[dim+1];
  for (int i = 0; i < dim+1; i++)
    den[i] = 1L;
    
  for (int i = 0; i < inequality_count; i++) {
    enu[0] = -1; // shift inequality
    for (int j = 0; j < dim; j++) 
      enu[j+1] = inequality[i*(dim)+j];            

    lrs_set_row(P, Q, i+1, enu, den, GE);
  }
     
  itomp(ONE, Q->Lcm[0]); // circumvent a bug in lrs 
    
  lrs_mp_matrix Lin;
  if (!lrs_getfirstbasis (&P, Q, &Lin, TRUE)) { // "TRUE" means no output 
    return false;
    //die_interior_point("can not find interior point."); 
  }
  else {
    if (Q->nredundcol!=0) 
      die_interior_point("polyhedral cone contains linearity space.\n");

    lrs_mp_vector output;
    output = lrs_alloc_mp_vector(Q->n);
    if (lrs_getsolution (P, Q, output, 0)) {
      for (int i = 0; i < dim; i++)
	interior[i] = mptodouble(output[i+1]); // /mptodouble(output[0]);
    }
    else {
      return false;
      // die_interior_point("can not find interior point.");
    }

    lrs_clear_mp_vector (output, Q->n);
    lrs_free_dic(P, Q);
    lrs_free_dat(Q);   
    lrs_close("interior_point finished"); 

    return true;
  }
}

bool compute_relative_interior_point_in_facet
(
 const int dim,
 const int inequality_count,
 const int* inequality,
 const int facet_index,
 double* interior
 )
{
  if (!lrs_init ("interior_point")) 
    die_interior_point("could not initialize lrs.\n");
 
  lrs_dic *P;
  lrs_dat *Q;
       
  Q = lrs_alloc_dat("LRS globals");
  if (Q == NULL) die_interior_point("Q == NULL.\n");
  Q->m = inequality_count;
  Q->n = dim+1;
  Q->verbose = FALSE;
  //Q->debug=TRUE;   
    
  P = lrs_alloc_dic(Q);
  if (P == NULL) die_interior_point("p == NULL.\n");
    
  long den[dim+1];
  long enu[dim+1];
  for (int i = 0; i < dim+1; i++)
    den[i] = 1L;
    
  for (int i = 0; i < inequality_count; i++) {
    if (i == facet_index) 
      enu[0] = 0;
    else
      enu[0] = -1; // shift inequality
    for (int j = 0; j < dim; j++) 
      enu[j+1] = inequality[i*(dim)+j];            

    if (i == facet_index)
      lrs_set_row(P, Q, i+1, enu, den, EQ);
    else
      lrs_set_row(P, Q, i+1, enu, den, GE);
  }
     
  itomp(ONE, Q->Lcm[0]); // circumvent a bug in lrs
    
  lrs_mp_matrix Lin;
  if (!lrs_getfirstbasis (&P, Q, &Lin, TRUE)) { // "TRUE" means no output 
    return false;
    //die_interior_point("can not find interior point."); 
  }
  else {
    if (Q->nredundcol!=0) 
      die_interior_point("polyhedral cone contains linearity space.\n");

    lrs_mp_vector output;
    output = lrs_alloc_mp_vector(Q->n);
    if (lrs_getsolution (P, Q, output, 0)) {
      for (int i = 0; i < dim; i++)
	interior[i] = mptodouble(output[i+1]); 
    }
    else {
      return false;
      // die_interior_point("can not find interior point.");
    }

    lrs_clear_mp_vector (output, Q->n);
    lrs_free_dic(P, Q);
    lrs_free_dat(Q);   
    lrs_close("interior_point finished"); 

    return true;
  }
}




