/* -*- Mode:C++; -*- */

// pacoop.cc

#include <iostream>
#include <NTL/mat_ZZ.h>

#include "QQ.h"
#include "rig_maxdet.h"
#include "covering.h"
#include "interior_point.h"

#define COUT_PRECISION 10

// The setting of this parameter can cause trouble!
#define BEYOND 0.00001 
#define DUALITY_GAP_FOR_CERTIFICATES 0.00000001
// Note: APPROX_PRECISION in rig_maxdet.cc is a critical parameter as well!

void die_pacoop(const char* last_words)
{
  std::cout <<"pacoop.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

// compute the inner product (trace(A*B)) of two symmetric dim x dim matrices A and B 
// given in as vec1, vec2 in lower triangular form starting at indices index1, index2
int compute_inner_product
(
 const int dim, 
 const int* vec1, 
 const int index1,
 const int* vec2, 
 const int index2)
{
  int prod = 0;
  for (int row = 0; row < dim; row++)
    for (int col = 0; col <= row; col++)
      if (col == row)
	prod += vec1[index1+POSLT(dim,row,col)] * 
	  vec2[index2+POSLT(dim,row,col)] ;
      else
	prod += 2*vec1[index1+POSLT(dim,row,col)] * 
	  vec2[index2+POSLT(dim,row,col)] ;
  return prod;
}

// The following is currently needed for option -e
// (works fine, but should be replaced in future versions)
//
// vec_is_already_known gets an integer array with index*dim entries
// and checks whether or not the vector is among the preceeding index-1 ones.
bool vec_is_already_known
(
 const int dim, 
 const int index, 
 const int* vec_list) 
{
  bool is_known = false;
  for (int i = 0; i < index && !is_known; i++) { 
    int j = 0;
    for (; j < dim && (vec_list[index*dim+j] == vec_list[i*dim+j]); j++);
    if (j == dim)
      is_known = true;
  }
  return is_known;
}

// ==========
// === pa ===
// === co ===
// === op ===
// ==========

int main(int argc, char** argv)
{
  int output_level = OUTPUT_LEVEL_COMMUNICATIVE;
  bool read_and_exclude_boundary_facets = false;
  bool certify_local_optimality = false;
  bool set_duality_gap = false; 
  bool read_interior_point = false;
  bool read_possible_min_vecs = false;
  bool set_maximum_iterations = false;
  bool quick_and_dirty = false;

  int max_iterations; // = 100;
  double duality_gap; // = 0.000000001; 

  int c;
  while ((c = getopt (argc, argv, "hbcd:eim:qvn")) != -1)
    switch (c)
      {
      case 'h':
	std::cout <<"Usage: pacoop [options] <file\n";
	std::cout <<"Options:\n";
	std::cout <<" -h   show this help\n";
	std::cout <<" -b   read and exclude boundary facets\n";
	std::cout <<" -c   certify local optimality\n";
	std::cout <<" -d x set duality gap to x\n";
	std::cout <<" -e   read possible minimal edges (vectors)\n";
	std::cout <<" -i   read interior point\n";
	std::cout <<" -m N set maximum iterations to N\n";
	std::cout <<" -q   quick and dirty (non-rigorous)\n";
	std::cout <<" -v   be verbose\n";
	std::cout <<" -n   be non-verbose\n";
	return 0;
	break;
      case 'b':
	read_and_exclude_boundary_facets = true;
	break;
      case 'c':
	certify_local_optimality = true;
	break;
      case 'd':
	duality_gap = atof(optarg);
	set_duality_gap = true; 	
	break;
      case 'e':
	read_possible_min_vecs = true; 
	break;
      case 'i':
	read_interior_point = true;
	break;
      case 'm':
	max_iterations = atoi(optarg);
	set_maximum_iterations = true; 
	break;
      case 'q':
	quick_and_dirty = true; 
	break;
      case 'v':
	output_level = OUTPUT_LEVEL_ALL;
	break;
      case 'n':
	output_level = OUTPUT_LEVEL_UNCOMMUNICATIVE;
	break;
      default:
	std::cout <<"pacoop: invalid option\n";
	std::cout <<"Try 'pacoop -h' for more information.\n";
	return 1;
      }

  if (quick_and_dirty && certify_local_optimality) {
    std::cout <<"pacoop: option -c has no effect if envoked with option -q\n";
    std::cout <<"Try 'pacoop -h' for more information.\n";
    return 1;
  }

  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"*** pa ***\n";
    std::cout <<"*** co ***\n";
    std::cout <<"*** op ***\n\n";
  }

  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"* READING INPUT, expecting\n";
    std::cout <<"- dim (int)\n";
    std::cout <<"- simplex_count (int)\n";
    std::cout <<"- simplex (int[simplex_count*dim*dim])\n";
    if (read_possible_min_vecs) {
      std::cout <<"- edge_count (int)\n";
      std::cout <<"- edges (int[edge_count*dim])\n";
    }
    std::cout <<"- basis_count (int)\n";
    std::cout <<"- basis (int[basis_count*dim*(dim+1)/2])\n";
    std::cout <<"- inequality_count (int)\n";
    if (read_and_exclude_boundary_facets)
      std::cout <<"- boundary_facets (bool[inequality_count])\n";
    std::cout <<"- inequality (int[inequality_count*dim])\n";
    if (read_interior_point)
      std::cout <<"- interior (double[dim])\n";
    if (!set_maximum_iterations)
      std::cout <<"- max_iterations (int)\n";
    if (!set_duality_gap)
      std::cout <<"- duality_gap (double)\n";
    std::cout << std::endl;
  }

  // 1. read input  
  int dim;
  std::cin >>dim;

  int simplex_count;
  std::cin >>simplex_count;
  int* simplex;
  simplex = (int*) calloc(simplex_count*dim*dim, sizeof(int));
  if (simplex == NULL) die_pacoop("simplex == NULL.\n");
  for (int i = 0; i < simplex_count*dim*dim; i++)
    std::cin >>simplex[i];

  int edge_count;
  int* edges;
  if (read_possible_min_vecs) {
    std::cin >>edge_count;
    edges = (int*) calloc(edge_count*dim, sizeof(int)); 
    if (edges == NULL) die_pacoop("edges == NULL.\n");
    for (int i = 0; i < edge_count*dim; i++)
      std::cin >>edges[i];
  }
  else {
    edge_count = (dim+1)*dim/2*simplex_count;
    // every simplex has (dim+1)*dim/2 edges
    // for simplicity they are written into an
    // extra array without doublets
    // (this piece of code is not so nice and should be replaced by using STL sets)
    edges = (int*) calloc(edge_count*dim, sizeof(int)); // in general too much mem.
    if (edges == NULL) die_pacoop("edges == NULL.\n");
    int edge_index = 0;
    for (int simplex_index = 0; simplex_index < simplex_count; simplex_index++) {
      for (int vert = 0; vert < dim; vert++) {
	// vert = vert-origin is an edge
	for (int coeff = 0; coeff < dim; coeff++)
	  edges[edge_index*dim+coeff] = simplex[(simplex_index*dim+vert)*dim+coeff];
	if (!vec_is_already_known(dim,edge_index,edges)) 
	  edge_index++;
	// vert-sec_vert is an edge
	for (int sec_vert = 0; sec_vert < vert; sec_vert++) {
	  for (int coeff = 0; coeff < dim; coeff++)
	    edges[edge_index*dim+coeff] = simplex[(simplex_index*dim+vert)*dim+coeff] - 
	      simplex[(simplex_index*dim+sec_vert)*dim+coeff];
	  if (!vec_is_already_known(dim,edge_index,edges))
	    edge_index++;
	}
      }
    }
    edge_count = edge_index;
  } 

  int basis_count;
  std::cin >>basis_count;
  int* basis;
  basis = (int*) calloc((basis_count+1)*dim*(dim+1)/2, sizeof(int));
  if (basis == NULL) die_pacoop("basis == NULL.\n");
  for (int i = 0; i < basis_count*dim*(dim+1)/2; i++) 
    std::cin >>basis[i];
  for (int i = basis_count*dim*(dim+1)/2; i < (basis_count+1)*dim*(dim+1)/2; i++) 
    basis[i] = 0; // dummy for extra variable

  int inequality_count;
  std::cin >>inequality_count;

  bool* boundary_facets;
  boundary_facets = (bool*) calloc(inequality_count, sizeof(bool));
  if (boundary_facets == NULL) die_pacoop("boundary_facets == NULL.\n");
  if (read_and_exclude_boundary_facets) {
    for (int i = 0; i < inequality_count; i++) 
      std::cin >>boundary_facets[i];
  }
  else {
    for (int i = 0; i < inequality_count; i++) 
      boundary_facets[i] = false;
  }
  
  int* inequality;
  if (!certify_local_optimality) 
    inequality = (int*) calloc((inequality_count+edge_count)*(basis_count+1), sizeof(int));
  else // for computation of certificates an extra inequality is used (see 3.B.II.)
    inequality = (int*) calloc((inequality_count+edge_count+1)*(basis_count+1), sizeof(int));
  if (inequality == NULL) die_pacoop("inequality == NULL.\n");
  for (int i = 0; i < inequality_count; i++) {
    for (int j = 0; j < basis_count; j++) {
      std::cin >>inequality[i*(basis_count+1)+j];
    }
    inequality[i*(basis_count+1)+basis_count] = 0;
  }
  // Additional ineq.: Q[v] = < x_1B_1 + ... + x_mB_m , [v] >
  //  (for edges v)         = < B_1, [v] > x_1 + ... + < B_m, [v] > x_m    >=    x_{m+1}
  for (int edge_index = 0; edge_index < edge_count; edge_index++) {
    int* bracket_v;
    bracket_v = (int*) calloc(dim*(dim+1)/2, sizeof(int));
    if (bracket_v == NULL) die_pacoop("bracket_v == NULL.\n");
    for (int row = 0; row < dim; row++)
      for (int col = 0; col <= row; col++) {
	bracket_v[POSLT(dim,row,col)] = edges[edge_index*dim+row]*edges[edge_index*dim+col];
      }
    for (int basis_index = 0; basis_index < basis_count; basis_index++) 
      inequality[(inequality_count+edge_index)*(basis_count+1)+basis_index] 
        = compute_inner_product(dim,basis,basis_index*dim*(dim+1)/2,bracket_v,0);
    inequality[(inequality_count+edge_index)*(basis_count+1)+basis_count] = -1;    
    free(bracket_v);
  }

  maxdet_data md;
  setup_maxdet_covering_constraints(dim, simplex_count, simplex, basis_count+1, basis, inequality_count + edge_count, inequality, md);  
  setup_maxdet_packing_covering_cost_func(dim, basis_count+1, md);

  double* interior;
  interior = (double*) calloc(basis_count+1, sizeof(double));
  if (interior == NULL) die_pacoop ("interior == NULL.\n");
  if (read_interior_point) {
    for (int i = 0; i < basis_count; i++) 
      std::cin >>interior[i];
    interior[basis_count] = 0.0;
  }

  if (!set_maximum_iterations)
      std::cin >>max_iterations;
  
  if (!set_duality_gap)
    std::cin >>duality_gap;

  // interior point is computed after input has been read, since lrs may manipulate std::cin
  if (!read_interior_point) {
    if (!compute_interior_point(basis_count+1,inequality_count+edge_count,inequality,interior))
      die_pacoop("Could not obtain interior point.\n");
    //normalize maximum diagonal entry to 1 (e.g. to ensure mu(interior)<=1)
    double max_diag = 0.0;
    for (int i = 0; i < dim; i++) {
      double i_diag_entry = 0.0;
      for (int basis_index = 0; basis_index < basis_count; basis_index++) {
	i_diag_entry += interior[basis_index] * double(basis[basis_index*dim*(dim+1)/2+POSLT(dim,i,i)]);
      }
      if (i_diag_entry > max_diag)
        max_diag = i_diag_entry;
    }
    if (max_diag == 0.0)
      die_pacoop("Obtaining interior point failed.\n");
    for (int i = 0; i < basis_count+1; i++)
      interior[i] = interior[i] / max_diag;  
  }

  std::cout.precision(COUT_PRECISION);
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    if (!read_possible_min_vecs) {
      std::cout <<"\n* computed possible minimal vectors\n"; 
      for (int edge_index = 0; edge_index < edge_count; edge_index++) {
	for (int coeff = 0; coeff < dim; coeff++)
	  std::cout <<edges[edge_index*dim+coeff] <<" ";
	std::cout <<std::endl;
      }
      std::cout <<std::endl;
    }
    if (!read_interior_point) {
      std::cout <<"\n* computed interior point\n";
      for (int i = 0; i < basis_count; i++) 
	std::cout <<interior[i] <<" ";
      std::cout <<std::endl;
    }
    std::cout <<std::endl;
  }
    
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout.precision(COUT_PRECISION);
    std::cout <<"* dimension = " <<dim <<"\n";
    std::cout <<"* simplex_count = " <<simplex_count <<"\n";
    std::cout <<"* edge_count = " <<edge_count <<"\n";
    std::cout <<"* basis_count = " <<basis_count <<"\n";
    std::cout <<"* inequality_count = " <<inequality_count <<"\n";
    std::cout <<"* max_iterations = " <<max_iterations <<"\n";
    std::cout <<"* duality_gap = " <<duality_gap <<"\n";
  }

  // 2. solve MAXDET quick and dirty or rigourously

  if (quick_and_dirty) {
    double lower_bound;
    double upper_bound;

    solve_maxdet_nonrigorously(md, interior, duality_gap, max_iterations, lower_bound, upper_bound, output_level-1);
   
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      std::cout <<"\n* gamma_lower_bound ~ " << 2.0/sqrt(-lower_bound) <<"\n";
      std::cout <<"* gamma_upper_bound ~ " << 2.0/sqrt(-upper_bound) <<"\n";
    }

    free_maxdet_covering_data(md);
    free(boundary_facets);
    free(interior);
    free(inequality);
    free(basis);
    free(edges);
    free(simplex);
    return 0;
  }

  // Continue rigourously
  NTL::vec_ZZ minimizer_approx;
  NTL::vec_ZZ dual_Z_approx;       
  NTL::vec_ZZ dual_W_approx;       
  maxdet_bound lower_bound;
  maxdet_bound upper_bound;

  solve_maxdet_rigorously(md, interior, duality_gap, max_iterations, minimizer_approx, dual_Z_approx, dual_W_approx, lower_bound, upper_bound, output_level);

  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    QQ lin_part_diff;
    QQ det_part_diff;
    lin_part_diff = upper_bound.lin_part - lower_bound.lin_part;
    det_part_diff = upper_bound.det_part / lower_bound.det_part;
	
    if ((lin_part_diff == NTL::to_ZZ(0)) && (det_part_diff == NTL::to_ZZ(1))) {
      if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
	if ((lin_part_diff == 0) && (det_part_diff == 1))  
	  std::cout << "* CERTIFIED OPTIMUM *\n";
	
	if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	  std::cout << "------ Proof -------\n"; 
	  std::cout <<md.m <<"\n\n";
	  
	  std::cout <<md.L <<"\n";
	  for (int i=0; i<md.L; i++)  
	    std::cout << md.F_blkszs[i] << " ";
	  std::cout << std::endl;
	  std::cout <<md.F << "\n\n";
	  
	  std::cout <<md.K <<"\n";
	  for (int i=0; i<md.K; i++)  
	    std::cout << md.G_blkszs[i] << " ";
	  std::cout << std::endl;
	  std::cout <<md.G << "\n\n";

	  std::cout <<md.c << "\n\n";
	  
	  std::cout <<minimizer_approx << "\n";
	  std::cout <<dual_Z_approx << "\n";
	  std::cout <<dual_W_approx << "\n";
	  std::cout << "------- qed --------\n"; 
	}
	else
	  std::cout << "* NOT A CERTIFIED OPTIMUM *\n\n";
      }
    }
  }
  
  // 3. obtain additional certificates
  
  bool lies_on_boundary = false;
  bool probably_on_facets = false;
  if (certify_local_optimality) {
    
    if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
      std::cout <<"* TRYING TO OBTAIN ADDITIONAL CERTIFICATES\n";
    }
    
    // 3.A. without linear constraints: (sec_upper_bound <= lower_bound)  implies "opt in bd(cone)"
    maxdet_data sec_md;
    setup_maxdet_covering_constraints(dim, simplex_count, simplex, basis_count+1, basis, edge_count, &inequality[inequality_count*(basis_count+1)], sec_md);
    setup_maxdet_packing_covering_cost_func(dim, basis_count+1, sec_md);
    NTL::vec_ZZ sec_minimizer_approx;
    NTL::vec_ZZ sec_dual_Z_approx;       
    NTL::vec_ZZ sec_dual_W_approx;       
    maxdet_bound sec_lower_bound;
    maxdet_bound sec_upper_bound;
    solve_maxdet_rigorously(sec_md, interior, duality_gap, max_iterations, sec_minimizer_approx, sec_dual_Z_approx, sec_dual_W_approx, sec_lower_bound, sec_upper_bound, output_level-1);
    free_maxdet_covering_data(sec_md);
    
    if (lower_bound > sec_upper_bound) {
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	std::cout <<"* CERTIFIED \"opt lies on boundary\"\n";
      }
      lies_on_boundary = true;
    }
    else { 
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	std::cout <<"* NO CERTIFICATE for \"opt lies on boundary\"\n";
      }
    }
    
    // 3.B. for each linear constraint (facet of cone), 
    //      determine minimum "on the other side", if it is not a boundary facet
    //      (sec_lower_bound > upper_bound) implies "opt not on facet"
    for (int ineq_index = 0; ineq_index < inequality_count; ineq_index++) 
      if (!boundary_facets[ineq_index]) {
	// B.I. construct sec_interior as interior point of facet and "beyond a little" 
	double* sec_interior;
	sec_interior = (double*) calloc(basis_count+1, sizeof(double));
	if (sec_interior == NULL) die_pacoop ("sec_interior == NULL.\n");
	
	if (!compute_relative_interior_point_in_facet(basis_count+1,inequality_count+edge_count,inequality,ineq_index,sec_interior))
	  die_pacoop("Could not obtain interior point.\n");
	
	double in_pr_ineq_ineq = 0.0;
	for (int i = 0; i < basis_count; i++) 
	  in_pr_ineq_ineq  += double(inequality[ineq_index*(basis_count+1) + i])*double(inequality[ineq_index*(basis_count+1) + i]);
	
	for (int i = 0; i < basis_count + 1; i++) 
	  sec_interior[i] = sec_interior[i] - BEYOND / sqrt(in_pr_ineq_ineq) * double(inequality[ineq_index*(basis_count+1) + i]);

	// normalize maximum diagonal entry to 1
	double max_diag = 0.0;
	for (int i = 0; i < dim; i++) {
	  double i_diag_entry = 0.0;
	  for (int basis_index = 0; basis_index < basis_count; basis_index++) {
	    i_diag_entry += sec_interior[basis_index] * double(basis[basis_index*dim*(dim+1)/2+POSLT(dim,i,i)]);
	  }
	  if (i_diag_entry > max_diag)
	    max_diag = i_diag_entry;
	}
	if (max_diag == 0.0)
	  die_pacoop("Obtaining interior point failed.\n");
	for (int i = 0; i < basis_count+1; i++)
	  sec_interior[i] = sec_interior[i] / max_diag;  

	// B.II. setup MAXDET with one inequality (switched)  ineq_index
	for (int basis_index = 0; basis_index < basis_count + 1; basis_index++) 
	  inequality[(inequality_count+edge_count)*(basis_count+1) + basis_index] 
            = -inequality[ineq_index*(basis_count+1) + basis_index];
	//inequality[(inequality_count+edge_count)*(basis_count+1)+basis_count] = 0;	
	
	setup_maxdet_covering_constraints(dim, simplex_count, simplex, basis_count+1, basis, edge_count+1,&inequality[inequality_count*(basis_count+1)], sec_md);
	setup_maxdet_packing_covering_cost_func(dim, basis_count+1, sec_md);
	solve_maxdet_rigorously(sec_md, sec_interior, DUALITY_GAP_FOR_CERTIFICATES, max_iterations, sec_minimizer_approx, sec_dual_Z_approx, sec_dual_W_approx, sec_lower_bound, sec_upper_bound, output_level-1);
	free_maxdet_covering_data(sec_md);
	free(sec_interior);
	
	if (sec_lower_bound > upper_bound) {
	  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
	    std::cout <<"* CERTIFIED \"opt does not lie on facet " <<ineq_index+1 <<"\"\n";
	  }
	}
	else {
	  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) { 
	    std::cout <<"* NO CERTIFICATE for \"opt not on facet " <<ineq_index+1 <<"\"\n";
	  }
	  probably_on_facets = true;
	}
      }
  }  
  
  // 4. summary 
  
  QQ lin_part_diff;
  QQ det_part_diff;
  lin_part_diff = upper_bound.lin_part - lower_bound.lin_part;
  det_part_diff = upper_bound.det_part / lower_bound.det_part;
  
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"\n";
    std::cout <<"***********\n";
    std::cout <<"* SUMMARY *\n";
    std::cout <<"***********\n";
    std::cout <<"\n";
  }
  
  if (certify_local_optimality) {
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      if (lies_on_boundary && !probably_on_facets)
	std::cout << "* CERTIFICATES CONTRADICT EACH OTHER (indicating approximation problems)\n";
      if (lies_on_boundary)
	std::cout << "* CERTIFICATE for optimum in bd(cone)\n";
      if (!probably_on_facets) {
	std::cout << "* CERTIFICATE for optimum in int(cone)\n";
	if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE && read_and_exclude_boundary_facets) {
	  std::cout << "  -- assuming boundary facets in input contain non-positive forms only --\n";
	}
      }
      if (probably_on_facets && !lies_on_boundary) {
	std::cout << "* NO CERTIFICATE for optimum in bd(cone) or optimum in int(cone)\n";
	std::cout << "  -- optimum might be exactly on boundary --\n";
      }
    }
  }
  
  minimizer_approx.SetLength(dim*(dim+1)/2+1); // trim to get rid of extra variable
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"* minimizer_approx = " <<minimizer_approx <<"\n";
  }
  
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"* dual_Z_approx = " <<dual_Z_approx <<"\n";
    std::cout <<"* dual_W_approx = " <<dual_W_approx <<"\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"* gamma_lower_bound";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<" = 2/sqrt(" <<-lower_bound.lin_part <<" - log(" <<lower_bound.det_part <<"))\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<" ~ " << 2.0/sqrt(-to_double(lower_bound.lin_part) - log(to_double(lower_bound.det_part))) <<"\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"* gamma_upper_bound";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<" = 2/sqrt(" <<-upper_bound.lin_part <<" - log(" <<upper_bound.det_part <<"))\n";
  }
  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {  
    std::cout <<" ~ " << 2.0/sqrt(-to_double(upper_bound.lin_part) - log(to_double(upper_bound.det_part))) <<"\n";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"* duality_gap";
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<" = " <<lin_part_diff <<" + log(" <<det_part_diff <<")\n";
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {  
    std::cout <<" ~ " <<to_double(lin_part_diff) + log(to_double(det_part_diff)) <<"\n" ;
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"* Thanks for your patience...\n";
  }
    
  // 5. cleaning

  free_maxdet_covering_data(md);
  free(boundary_facets);
  free(interior);
  free(inequality);
  free(basis);
  free(edges);
  free(simplex);
  
  return 0;
}

