/* -*- Mode:C++; -*- */

// rig_maxdet.cc
   
#include <iostream>
#include <vector>
#include <NTL/mat_ZZ.h>

#include "QQ.h"
#include "rig_maxdet.h"

#define MAX(x,y)  ((x) > (y) ? (x) : (y))

#define APPROX_PRECISION 20 // 10^-APPROX_PRECISION should be smaller than duality_gap when rigorousMAXDET is called
#define MAXDET_GAMMA 1000
#define MAXDET_RELTOL 0.0
#define MINABSTOL 1e-10 /* min absolute tolerance allowed */
#define MAX_LENGTH_CONTINUED_FRACTION 20
#define CONTINUED_FRACTION_ZERO 1e-20

// ============================
// === "traditional" maxdet ===
// ============================

//
// code and details are in maxdet_src.c
// 

extern "C" int maxdet(
 int m,                /* no of variables */
 int L,                /* no of blocks in F */
 double *F,            /* F_i's in packed storage */
 int *F_blkszs,        /* L-vector, dimensions of diagonal blocks of F */
 int K,                /* no of blocks in G */
 double *G,            /* G_i's in packed storage */
 int *G_blkszs,        /* K-vector, dimensions of diagonal blocks of G */
 double *c,            /* m-vector */
 double *x,            /* m-vector */
 double *Z,            /* block diagonal matrix in packed storage */
 double *W,            /* block diagonal matrix in packed storage */
 double *ul,           /* ul[0] = pr. obj, ul[1] = du. obj */
 double *hist,         /* history, 3-by-NTiter matrix */
 double gamma,         /* > 0 */
 double abstol,        /* absolute accuracy */
 double reltol,        /* relative accuracy */
 int *NTiters,         /* on entry: maximum number of (total) Newton iters,
                        * on exit: number of Newton iterations taken */
 double *work,         /* (double) work array */
 int lwork,            /* size of work */
 int *iwork,           /* (int) work array */
 int *info,            /* status on termination */
 int output_level      /* if >= OUTPUT_LEVEL_COMMUNICATIVE, */
                       /* maxdet provides some output */
 );

// =======================
// === rigorous maxdet ===
// =======================

void die_rig_maxdet(char* last_words)
{
  std::cout <<"rig_maxdet.cc: " <<last_words;
  exit(EXIT_FAILURE);
}

NTL::ZZ global_scaling_factor = power(NTL::to_ZZ(10), APPROX_PRECISION);

void convert_double_to_scaled_ZZ (const double from, NTL::ZZ& to)
{
  NTL::vec_ZZ a;
  a.SetLength(MAX_LENGTH_CONTINUED_FRACTION);
  double r = from;
  
  int n = 0;
  bool is_zero = false;
  while (n < MAX_LENGTH_CONTINUED_FRACTION && !is_zero) {
    a[n] = NTL::to_ZZ(floor(r + 0.5));
    if (fabs(r - NTL::to_double(a[n])) < CONTINUED_FRACTION_ZERO) {
      is_zero = true;
    }
    else {
      r = 1/(r - NTL::to_double(a[n]));
    }
    n++;
  }

  QQ p(a[n-1]);
  QQ q2(1);
  int i = n - 2;
  while (i >= 0) {
    QQ q1(a[i]);
    p = q1 + q2/p;
    i--;
  }

  to = (Num(p) * global_scaling_factor)/Den(p);
}

void output_maxdet_data(const maxdet_data& md) {
  std::cout <<"- MAXDET DATA\n";
  std::cout <<"- m = " <<md.m <<"\n";
  std::cout <<"- L = " <<md.L <<"\n";
  std::cout <<"- F_blkszs = "; 
  for (int i = 0; i < md.L; i++)
    std::cout <<md.F_blkszs[i] <<" ";
  std::cout <<"\n";
  std::cout <<"- F = " <<md.F <<"\n";
  std::cout <<"- K = " <<md.K <<"\n";
  std::cout <<"- G_blkszs = "; 
  for (int i = 0; i < md.K; i++)
    std::cout <<md.G_blkszs[i] <<" ";
  std::cout <<"\n";
  std::cout <<"- G = " <<md.G <<"\n";
  std::cout <<"- c = " <<md.c <<"\n";
}

// reduce_vector divides the entries of vec by the gcd of all entries 
void reduce_vector(NTL::vec_ZZ& vec)
{
  if (vec.length() >= 2) {
    NTL::ZZ gcd;
    GCD(gcd, vec[0], vec[1]);
    for (int i = 2; i < vec.length(); i++)
      GCD(gcd, gcd, vec[i]);
    for (int i = 0; i < vec.length(); i++)
      vec[i] = vec[i] / gcd;
  }
}

// test, whether matrix is positive definite (definite = true)
// or whether it is positive semidefinite (definite = false)

bool is_positive_semidefinite
(
 const NTL::mat_ZZ& matrix, 
 const bool definite
)
{
  for (int minor_size = 1; minor_size <= matrix.NumRows(); minor_size++) {
    NTL::mat_ZZ zz_minor;
    zz_minor.SetDims(minor_size, minor_size);
    for (int row = 0; row < minor_size; row++)
      for (int col = 0; col < minor_size; col++)
	zz_minor[row][col] = matrix[row][col];
    NTL::ZZ det;
    determinant(det, zz_minor, true);
    if (det < 0 || (det == 0 && definite)) return false;
  }
  return true;
}

// same as above for matrices in upper triangular form. 
// note: first coefficient of matrix is skipped

bool is_positive_semidefinite
(
 const int dim,
 const NTL::vec_ZZ& matrix,
 const bool definite
)
{
  for (int minor_size = 1; minor_size <= dim; minor_size++) {
    NTL::mat_ZZ zz_minor;
    zz_minor.SetDims(minor_size, minor_size);
    for (int row = 0; row < minor_size; row++)
      for (int col = row; col < minor_size; col++) {
	zz_minor[row][col] = matrix[POSUT(dim,row,col)+1];
	zz_minor[col][row] = zz_minor[row][col];
      }
    NTL::ZZ det;
    determinant(det, zz_minor, true);
    if (det < 0 || (det == 0 && definite)) return false;
  }
  return true;
}

// same as above for matrices in upper triangular
// block diagonal form, note: first coefficient of matrix is skipped

bool is_positive_semidefinite
(
 const int block_count,
 const int* block_szs,
 const NTL::vec_ZZ& block_matrix,
 const bool definite
)
{
  int block_sz = 0;
  for (int block = 0; block < block_count; block++) {
    for (int minor_size = 1; minor_size <= block_szs[block]; minor_size++) {
      NTL::mat_ZZ zz_minor;
      zz_minor.SetDims(minor_size, minor_size);
      for (int row = 0; row < minor_size; row++)
	for (int col = row; col < minor_size; col++) {
	  zz_minor[row][col] = block_matrix[1 + block_sz + POSUT(block_szs[block],row,col)];
	  zz_minor[col][row] = zz_minor[row][col];
	}
      NTL::ZZ det;
      determinant(det, zz_minor, true);
      if (det < 0 || (det == 0 && definite)) return false;
    }
    block_sz += (block_szs[block] + 1) * block_szs[block] / 2;
  }
  return true;
}

NTL::ZZ compute_tr_inner_product
(
 const int dim,
 const NTL::vec_ZZ& vec1, 
 const int index1,
 const NTL::vec_ZZ& vec2,
 const int index2
)
{
  NTL::ZZ prod;
  prod = 0;
  for (int row = 0; row < dim; row++)
    for (int col = row; col < dim; col++)
      if (row == col)
	prod += vec1[index1 + POSUT(dim,row,col)] * vec2[index2 + POSUT(dim,row,col)];
      else
	prod += 2 * vec1[index1 + POSUT(dim,row,col)] * vec2[index2 + POSUT(dim,row,col)];
  return prod;
}

NTL::ZZ compute_tr_inner_product
(
 const int block_count,
 const int* block_szs,
 const NTL::vec_ZZ& vec1, 
 const int index1,
 const NTL::vec_ZZ& vec2,
 const int index2
 )
{
  NTL::ZZ prod;
  prod = 0;
  int block_sz = 0;
  for (int block = 0; block < block_count; block++) {
    prod += compute_tr_inner_product(block_szs[block], vec1, index1 + block_sz, vec2, index2 + block_sz);
    block_sz += (block_szs[block] + 1)*block_szs[block]/2;
  }
  return prod;
}

// project_onto_F_space 
// ====================
//
// project from onto to in { Y : Tr F_i Y = c_i - Tr G_i WApprox, i = 1..m }
// from and to can be the same without conflicts

void project_onto_F_space 
(
 const maxdet_data& md, 
 const NTL::vec_ZZ& W_approx, 
 const NTL::vec_ZZ& from, 
 NTL::vec_ZZ& to
 ) 
{  
  // check input
  int F_sz = (md.F.length() - 1)/(md.m + 1);
  if(from.length() != F_sz+1)
    die_rig_maxdet("from.length() != F_sz+1\n");

  NTL::mat_ZZ gram_matrix;
  gram_matrix.SetDims(md.m, md.m);
  for (int row = 0; row < md.m; row++)
    for (int col = 0; col <= row; col++) {
      gram_matrix[row][col] = 
	compute_tr_inner_product(md.L,md.F_blkszs,md.F, 1+(row+1)*F_sz,md.F,1+(col+1)*F_sz);
      gram_matrix[col][row] = gram_matrix[row][col];
    }
  NTL::ZZ det;
  determinant(det, gram_matrix, true);
  if (det == 0) {
    die_rig_maxdet("pos. def. blocks of F_1,..,F_m are not linear independent!\nCan not handle this case...\n");
  }

  // ok, we can do it...
  // We assume Y = sum_j alpha_j F_j and obtain alpha_j as solutions of
  // sum_j alpha_j Tr F_i F_j = c_i - Tr G_i W_approx - Tr F_i from 
  // the denominators c,g,w,f,z of c,G,W_approx,F,from 
  // (here and in the sequel) have to be brought into the linear system
  // by multiplying both sides with c*f^2*g*z*w
  gram_matrix = (md.c[0] * md.G[0] * W_approx[0] * from[0]) * gram_matrix;
    
  // obtain right hand side of scaled linear system
  NTL::vec_ZZ rhs;
  rhs.SetLength(md.m);
  int G_sz = (md.G.length()-1)/(md.m+1);
  for (int row = 0; row < md.m; row++) {
    rhs[row] = (md.F[0] * md.F[0] * md.G[0] * W_approx[0] * from[0]) * md.c[row+1];
    rhs[row] += -(md.c[0] * md.F[0] * md.F[0] * from[0]) * 
      compute_tr_inner_product(md.K,md.G_blkszs,md.G,1+(row+1)*G_sz,W_approx,1);
    rhs[row] += -(md.c[0] * md.F[0] * md.G[0] * W_approx[0]) * 
      compute_tr_inner_product(md.L,md.F_blkszs,md.F,1+(row+1)*F_sz,from,1);
  }
  
  // solve linear system and obtain to from it
  // (one has to deal with the denominators of from, F_j and alpha (multiplier))
  NTL::ZZ multiplier; // in solve multiplier is set to determinant of gram_matrix 
  NTL::vec_ZZ alpha;
  solve(multiplier, alpha, gram_matrix, rhs, true);

  NTL::vec_ZZ tmp;
  tmp.SetLength(F_sz+1);
  tmp[0] = from[0] * md.F[0] * multiplier;       
  int F_block_sz = 0;
  for (int block = 0; block < md.L; block++) {
    for (int index = 0; index < (md.F_blkszs[block] + 1) * md.F_blkszs[block] / 2; index++) {
      tmp[1 + F_block_sz + index] = (md.F[0] * multiplier) * from[1 + F_block_sz + index];
      for (int row = 0; row < md.m; row++) 
	tmp[1 + F_block_sz + index] += from[0] * (alpha[row] * md.F[1 + (row+1)*F_sz + F_block_sz + index]);
    }
    F_block_sz += (md.F_blkszs[block] + 1) * md.F_blkszs[block] / 2;
  }

  reduce_vector(tmp);
  to = tmp;
}

// project_onto_G_space
// ====================
//
// project from onto to in { X : Tr G_i X = c_i - Tr F_i Z_approx , i=1..m }
// from and to can be the same without conflicts

void project_onto_G_space 
(
 const maxdet_data& md, 
 const NTL::vec_ZZ& Z_approx, 
 const NTL::vec_ZZ& from, 
 NTL::vec_ZZ& to
 ) 
{  
  // check input
  int G_sz = (md.G.length()-1)/(md.m+1);
  if (from.length() != G_sz+1)
    die_rig_maxdet("from.length() != G_sz+1\n");

  NTL::mat_ZZ gram_matrix;
  gram_matrix.SetDims(md.m, md.m);
  for (int row = 0; row < md.m; row++)
    for (int col = 0; col <= row; col++) {
      gram_matrix[row][col] = compute_tr_inner_product(md.K,md.G_blkszs,md.G, 1+(row+1)*G_sz,md.G,1+(col+1)*G_sz);
      gram_matrix[col][row] = gram_matrix[row][col];
    }
  NTL::ZZ det;
  determinant(det, gram_matrix, true);
  if (det == 0) {
    die_rig_maxdet("pos. def. blocks of G_1,..,G_m are not linear independent!\nCan not handle this case...\n");
  }

  // ok, we can do it...
  // We assume X = sum_j alpha_j G_j and obtain alpha_j as solutions of
  // sum_j alpha_j Tr G_i G_j = c_i - Tr F_i Z_approx - Tr G_i from 
  // the denominators c,g,z,f,w of c,G,Z_approx,F,from 
  // (here and in the sequel) have to be brought into the linear system
  // by multiplying both sides with c*g^2*f*z*w
  gram_matrix = (md.c[0] * md.F[0] * Z_approx[0] * from[0]) * gram_matrix;
    
  // obtain right hand side of scaled linear system
  NTL::vec_ZZ rhs;
  rhs.SetLength(md.m);
  int F_sz = (md.F.length()-1)/(md.m+1);
  for (int row = 0; row < md.m; row++) {
    rhs[row] = (md.G[0] * md.G[0] * md.F[0] * Z_approx[0] * from[0]) * md.c[row+1];
    rhs[row] += -(md.c[0] * md.G[0] * md.G[0] * from[0]) * 
      compute_tr_inner_product(md.L,md.F_blkszs,md.F,1+(row+1)*F_sz,Z_approx,1);
    rhs[row] += -(md.c[0] * md.G[0] * md.F[0] * Z_approx[0]) *
      compute_tr_inner_product(md.K,md.G_blkszs,md.G,1+(row+1)*G_sz,from,1);
  }
  
  // solve linear system and obtain to from it
  // (one has to deal with the denominators of from, G_j and alpha (multiplier))
  NTL::ZZ multiplier; // in solve multiplier is set to determinant of gram_matrix 
  NTL::vec_ZZ alpha;
  solve(multiplier, alpha, gram_matrix, rhs, true);

  NTL::vec_ZZ tmp;
  tmp.SetLength(G_sz+1);
  tmp[0] = from[0] * md.G[0] * multiplier;       
  int G_block_sz = 0;
  for (int block = 0; block < md.K; block++) {
    for (int index = 0; index < (md.G_blkszs[block] + 1) * md.G_blkszs[block] / 2; index++) {
      tmp[1 + G_block_sz + index] = (md.G[0] * multiplier) * from[1 + G_block_sz + index];
      for (int row = 0; row < md.m; row++) 
	tmp[1 + G_block_sz + index] += from[0] * (alpha[row] * md.G[1 + (row+1)*G_sz + G_block_sz + index]);
    }
    G_block_sz += (md.G_blkszs[block] + 1) * md.G_blkszs[block] / 2;
  }
  reduce_vector(tmp);
  to = tmp;  
}

// check_primal_solution
// =====================

bool check_primal_solution
(
 const maxdet_data& md, 
 const NTL::vec_ZZ& minimizer_approx,
 maxdet_bound& upper_bound,
 int output_level
 )
{
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"\n";
    std::cout <<"- CHECKING PRIMAL SOLUTION\n";
  }
 
  // 1. compute blocks of G(x) and test if they are all positive definite

  int G_sz = (md.G.length() - 1)/(md.m + 1);
  int G_block_sz = 0;  
  int G_dim = 0;
  QQ G_det = 1;
  for (int block = 0; block < md.K; block++) {
    NTL::mat_ZZ G_of_X;
    G_of_X.SetDims(md.G_blkszs[block], md.G_blkszs[block]);
    for (int var_index = 0; var_index <= md.m; var_index ++) {
      for (int row = 0; row < md.G_blkszs[block]; row++)
	for (int col = row; col < md.G_blkszs[block]; col++)
	  G_of_X[row][col] += minimizer_approx[var_index] * 
	    md.G[1 + var_index*G_sz + G_block_sz + POSUT(md.G_blkszs[block],row,col)];
    }
    for (int row = 0; row < md.G_blkszs[block]; row++)
      for (int col = 0; col < row; col++)
	G_of_X[row][col] = G_of_X[col][row];

    G_block_sz += (md.G_blkszs[block] + 1) * md.G_blkszs[block] / 2;
    G_dim += md.G_blkszs[block];
 
    if (!is_positive_semidefinite(G_of_X, true))
      return false;

    NTL::ZZ det;  
    determinant(det, G_of_X, true);
    G_det = G_det*det;
  }
  G_det = G_det/power(minimizer_approx[0]*md.G[0],G_dim);
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- G(x) is positive definite.\n";

  // 2. compute blocks of F(x) and test if they are all positive semidefinite

  int F_sz = (md.F.length() - 1)/(md.m + 1);
  int F_block_sz = 0;  
  for (int block = 0; block < md.L; block++) {
    NTL::mat_ZZ F_of_X;
    F_of_X.SetDims(md.F_blkszs[block],md.F_blkszs[block]);
    for (int var_index = 0; var_index <= md.m; var_index ++) {
      for (int row = 0; row < md.F_blkszs[block]; row++)
	for (int col = row; col < md.F_blkszs[block]; col++) 
	  F_of_X[row][col] += minimizer_approx[var_index] *
	    md.F[1 + var_index*F_sz + F_block_sz + POSUT(md.F_blkszs[block],row,col)];
    }
    for (int row = 0; row < md.F_blkszs[block]; row++)
      for (int col = 0; col < row; col++)
	F_of_X[row][col] = F_of_X[col][row];

    F_block_sz += (md.F_blkszs[block]+1)*md.F_blkszs[block]/2;

    if (!is_positive_semidefinite(F_of_X, false)) {
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
	std::cout <<"- F(x) is not positive semidefinite in block " <<block + 1 <<".\n";
      return false;
    }
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- F(x) is positive semidefinite.\n";

  // 3. if everything ok then compute upper_bound for min c^t x - log(det G)

  upper_bound.det_part = QQ(1)/G_det;
  NTL::ZZ inner_prod;
  inner_prod = 0;
  for (int i = 1; i <= md.m; i++)
    inner_prod += md.c[i]*minimizer_approx[i];
  upper_bound.lin_part = QQ(inner_prod, md.c[0]*minimizer_approx[0]);

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- upper_bound";
  if (output_level >= OUTPUT_LEVEL_ALL)
    std::cout <<" = " <<upper_bound.lin_part <<" + log(" <<upper_bound.det_part <<")";
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<" ~ " <<to_double(upper_bound.lin_part) + log(to_double(upper_bound.det_part)) <<"\n";

  return true;
}

// check_dual_solution
// ===================

bool check_dual_solution
(
 const maxdet_data& md,
 const NTL::vec_ZZ& dual_Z_approx,
 const NTL::vec_ZZ& dual_W_approx,
 maxdet_bound& lower_bound,
 const int output_level
 )
{
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"\n";
    std::cout <<"- CHECKING DUAL SOLUTION\n";
  }

  // 1. check if dual_{W,Z}_approx is in { (X,Y) : Tr G_i X + Tr F_i Y = c_i }
  // (here one has to deal with the denominators) 

  int G_sz = (md.G.length() - 1)/(md.m + 1);
  int F_sz = (md.F.length() - 1)/(md.m + 1);
  for (int row = 0; row < md.m; row++) {
    if ((md.F[0]*dual_Z_approx[0]*md.c[0]) * 
	compute_tr_inner_product(md.K,md.G_blkszs,md.G,1+(row+1)*G_sz,dual_W_approx,1) + 
	(md.G[0]*dual_W_approx[0]*md.c[0]) * 
	compute_tr_inner_product(md.L,md.F_blkszs,md.F,1+(row+1)*F_sz,dual_Z_approx,1)
	!= (md.G[0]*dual_W_approx[0]*md.F[0]*dual_Z_approx[0]) * md.c[row+1] ) {
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
	std::cout <<"- Tr G_" <<row+1 << " W + Tr F_" <<row+1 << " Z <> c_" <<row+1 << ".\n";
      return false;
    }
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- Tr G_i W + Tr F_i Z = c_i.\n";

  // 2. check if blocks of dual_W_approx are positive definite

  int G_block_sz = 0;
  int G_dim = 0; 
  QQ W_det = 1;  
  for (int block = 0; block < md.K; block++) {
    NTL::mat_ZZ dual_W_approx_block;
    dual_W_approx_block.SetDims(md.G_blkszs[block],md.G_blkszs[block]);
    for (int row = 0; row < md.G_blkszs[block]; row++)
      for (int col = 0; col < md.G_blkszs[block]; col++) {
        if (col >= row)
	  dual_W_approx_block[row][col] = dual_W_approx[1 + G_block_sz + POSUT(md.G_blkszs[block],row,col)];
	else
	  dual_W_approx_block[row][col] = dual_W_approx[1 + G_block_sz + POSUT(md.G_blkszs[block],col,row)];
      }
    if (!is_positive_semidefinite(dual_W_approx_block, true)) {
      if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
	std::cout <<"- W is not positive definite.\n";
      return false;
    }
    
    G_block_sz += (md.G_blkszs[block] + 1) * md.G_blkszs[block] / 2;	
    G_dim += md.G_blkszs[block];

    NTL::ZZ det;  
    determinant(det, dual_W_approx_block, true);
    W_det = W_det * det;
  }
  W_det = W_det / power(dual_W_approx[0], G_dim);
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- W is positive definite.\n";
  
  // 3. check if dual_Z_approx is positive semidefinite

  if (!is_positive_semidefinite(md.L, md.F_blkszs, dual_Z_approx, false)) {
    if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
      std::cout <<"- Z is not positive semidefinite.\n";
    }
    return false;
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- Z is positive semidefinite.\n";

  // 4. If everything ok, then compute lower_bound max G_dim - Tr G_0 W - Tr F_0 Z + log(det W)

  lower_bound.det_part = W_det;
  lower_bound.lin_part = QQ(G_dim);
  lower_bound.lin_part = lower_bound.lin_part - 
    QQ(compute_tr_inner_product(md.K,md.G_blkszs,md.G,1,dual_W_approx,1),
       (md.G[0] * dual_W_approx[0]));
  lower_bound.lin_part = lower_bound.lin_part - 
    QQ(compute_tr_inner_product(md.L,md.F_blkszs,md.F,1,dual_Z_approx,1),
       (md.F[0] * dual_Z_approx[0]));

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- lower_bound ";
  if (output_level >= OUTPUT_LEVEL_ALL)
    std::cout <<" = " <<lower_bound.lin_part << " + log(" <<lower_bound.det_part <<")";
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<" ~ " <<to_double(lower_bound.lin_part) + log(to_double(lower_bound.det_part)) <<"\n";

  return true;
}


// solve_maxdet_rigorously
// =======================

bool solve_maxdet_rigorously 
(
 const maxdet_data& md,
 const double* interior_point,
 const double duality_gap,
 const int max_iterations,
 NTL::vec_ZZ& minimizer_approx,
 NTL::vec_ZZ& dual_Z_approx,
 NTL::vec_ZZ& dual_W_approx,
 maxdet_bound& lower_bound,
 maxdet_bound& upper_bound,
 const int output_level
 )
{
  // 0. first do some checking on the involved parameters 
  if (duality_gap < MINABSTOL)
    std::cerr<<"\nNOTE: duality_gap < MINABSTOL = " <<MINABSTOL <<"\n";
  if (NTL::to_double(global_scaling_factor) < 1/duality_gap) 
    die_rig_maxdet("\nduality_gap is to small for APPROX_PRECISON.\n");

  // 1. solve "traditional MAXDET" with double
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"\n";
    std::cout <<"- SOLVING MAXDET PROBLEM\n";
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    output_maxdet_data(md);
    std::cout <<"- interior_point: ";
    for (int i = 0; i < md.m; i++)
      std::cout << interior_point[i] <<" ";
    std::cout <<"\n";
    std::cout <<"- duality_gap: " <<duality_gap <<"\n";  
    std::cout <<"- max_iterations: " <<max_iterations <<"\n"; 
  }

  // compute size of work space (see maxdet_src.c for explanation)
  int G_sz = (md.G.length() - 1)/(md.m + 1);
  int max_l = 0;
  int l = 0;
  for (int i = 0; i < md.K; i++) {
    l = l + md.G_blkszs[i];
    if (max_l < md.G_blkszs[i])
      max_l = md.G_blkszs[i];
  }
  int F_sz = (md.F.length() - 1)/(md.m + 1);
  int max_n = 0;
  int n = 0;
  for (int i = 0; i < md.L; i++) {
    n = n + md.F_blkszs[i];
    if (max_n < md.F_blkszs[i])
      max_n = md.F_blkszs[i];
  }
  int sz = F_sz + G_sz;
  int M_upsz = md.m * md.m;
  int NB = 32; 
  int lwork = (2*md.m+5)*sz + 2*(n+l) + 
    MAX(md.m+sz*NB,MAX(3*(md.m+M_upsz+MAX(G_sz,F_sz)),
		       MAX(3*(MAX(max_l,max_n)+MAX(G_sz,F_sz)),
		           MAX(G_sz+3*max_l,F_sz+3*max_n))));

  // define G
  double* G;
  G = (double*) calloc(md.G.length() - 1, sizeof(double));
  if (G == NULL) die_rig_maxdet("G == NULL.\n");
  for (int i = 0; i < md.G.length() - 1; i++)
    G[i] = NTL::to_double(md.G[i+1])/NTL::to_double(md.G[0]);
  
  // define F
  double *F;
  F = (double*) calloc(md.F.length() - 1, sizeof(double));
  if (F == NULL) die_rig_maxdet("F == NULL.\n");
  for (int i = 0; i < md.F.length() - 1; i++)
    F[i] = NTL::to_double(md.F[i+1])/NTL::to_double(md.F[0]);
  
  // define c
  double* c;
  c = (double*) calloc(md.c.length() - 1, sizeof(double));
  if (c == NULL) die_rig_maxdet("c == NULL.\n");
  for (int i = 0; i < md.c.length() - 1; i++)
    c[i] = NTL::to_double(md.c[i+1]) / NTL::to_double(md.c[0]);
  
  // define other parameters
  double ul[2];
  double* hist;
  hist = (double*) calloc(3*max_iterations, sizeof(double));
  if (hist == NULL) die_rig_maxdet("hist == NULL.\n");
  double gamma = MAXDET_GAMMA;
  double reltol = MAXDET_RELTOL;
  double* work;
  work = (double*) calloc(lwork, sizeof(double));
  if (work == NULL) die_rig_maxdet("work == NULL.\n");
  int iwork[3];
  int info; 

  double* minimizer;
  minimizer = (double*) calloc(md.m, sizeof(double));
  if (minimizer == NULL) die_rig_maxdet("minimizer == NULL\n.");
  for (int i = 0; i < md.m; i++)
    minimizer[i] = interior_point[i];   
 
  double *dual_Z;
  dual_Z = (double*) calloc(md.F.length() - 1, sizeof(double));
  if (dual_Z == NULL) die_rig_maxdet("dual_Z == NULL\n.");
  
  double *dual_W;
  dual_W = (double*) calloc(md.G.length() - 1, sizeof(double));
  if (dual_W == NULL) die_rig_maxdet("dual_W == NULL\n.");

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- starting MAXDET (workspace = " <<lwork <<")\n";

  maxdet(md.m, md.L, F, md.F_blkszs, md.K, G, md.G_blkszs, c,
 	 minimizer, dual_Z, dual_W, ul, hist, gamma, duality_gap,
 	 reltol, const_cast<int*>(&max_iterations), work, lwork, iwork, &info,
	 output_level);

  free(work);
  free(hist);
  free(c);
  free(G);
  free(F);

  if (info < 0) die_rig_maxdet("info < 0.\n");
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"- MAXDET stopped because ";
    switch (info) {
    case 1: std::cout <<"max_iterations exceeded (increase might help).\n"; break;
    case 2: std::cout <<"duality_gap reached.\n"; break;
    case 3: std::cout <<"reltol reached.\n"; break;
    }
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"- primal value = " <<ul[0] <<"\n";
    std::cout <<"- dual value   = " <<ul[1] <<"\n";
    std::cout <<"- minimizer = ";
    for (int i = 0; i < md.m; i++) 
      std::cout <<minimizer[i] <<" ";
    std::cout <<"\n";
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"- dual_Z = ";
    for (int i = 0; i < F_sz; i++)
      std::cout <<dual_Z[i] <<" ";
    std::cout <<"\n";
    std::cout <<"- dual_W = ";
    for (int i = 0; i < G_sz; i++) 
      std::cout <<dual_W[i] <<" ";
    std::cout <<"\n";
  }
  
  // 2. obtain minimizer_approx
  minimizer_approx.SetLength(md.m+1);
  minimizer_approx[0] = global_scaling_factor;
  for (int i = 0; i < md.m; i++)
    convert_double_to_scaled_ZZ(minimizer[i], minimizer_approx[i+1]);
  reduce_vector(minimizer_approx);
  free(minimizer);

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- minimizer_approx = " <<minimizer_approx <<"\n";

  // 3. obtain dual_W_approx
  dual_W_approx.SetLength(G_sz + 1); 
  dual_W_approx[0] = global_scaling_factor;
  for (int i = 0; i < G_sz; i++) 
    convert_double_to_scaled_ZZ(dual_W[i], dual_W_approx[i+1]);
  reduce_vector(dual_W_approx); 
  free(dual_W);

  // 4. obtain semidefinite dual_Z_approx
  NTL::vec_ZZ first_Z_approx;
  first_Z_approx.SetLength(F_sz+1); 
  first_Z_approx[0] = global_scaling_factor;
  for (int i = 0; i < F_sz; i++)
    convert_double_to_scaled_ZZ(dual_Z[i], first_Z_approx[i+1]);
  reduce_vector(first_Z_approx);

  project_onto_F_space(md, dual_W_approx, first_Z_approx, dual_Z_approx);
  free(dual_Z);

  if (!is_positive_semidefinite(md.L, md.F_blkszs, dual_Z_approx, false)) {
    dual_Z_approx = first_Z_approx;
    project_onto_G_space(md, dual_Z_approx, dual_W_approx, dual_W_approx); 
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"- dual_W_approx = " <<dual_W_approx <<"\n";
    std::cout <<"- dual_Z_approx = " <<dual_Z_approx <<"\n";
  }

  // 5. test primal feasibility and obtain upper_bound
  if (!check_primal_solution(md, minimizer_approx, upper_bound, output_level))
    return false;
    
  // 6. test dual feasibility and obtain lower_bound
  if (!check_dual_solution(md, dual_Z_approx, dual_W_approx, lower_bound,output_level)) 
    return false;

  // 7. compute duality gap
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    QQ lin_part_diff = upper_bound.lin_part - lower_bound.lin_part;
    QQ det_part_diff = upper_bound.det_part / lower_bound.det_part;
    std::cout <<"- duality_gap";
    if (output_level >= OUTPUT_LEVEL_ALL)      
      std::cout <<" = " <<lin_part_diff <<" + log(" <<det_part_diff <<")";
    std::cout <<" ~ " <<to_double(lin_part_diff) + log(to_double(det_part_diff)) <<"\n\n";
  }

  return true;
}


// solve_maxdet_nonrigorously
// ==========================

void solve_maxdet_nonrigorously 
(
 const maxdet_data& md,
 const double* interior_point,
 const double duality_gap,
 const int max_iterations,
 double& lower_bound,
 double& upper_bound,
 const int output_level
 )
{
  // 0. first do some checking on the involved parameters 
  if (duality_gap < MINABSTOL)
    std::cerr<<"\nNOTE: duality_gap < MINABSTOL = " <<MINABSTOL <<"\n";
  if (NTL::to_double(global_scaling_factor) < 1/duality_gap) 
    die_rig_maxdet("\nduality_gap is to small for APPROX_PRECISON.\n");
  
  // 1. solve "traditional MAXDET" with double
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"\n";
    std::cout <<"- SOLVING MAXDET PROBLEM\n";
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    output_maxdet_data(md);
    std::cout <<"- interior_point: ";
    for (int i = 0; i < md.m; i++)
      std::cout << interior_point[i] <<" ";
    std::cout <<"\n";
    std::cout <<"- duality_gap: " <<duality_gap <<"\n";  
    std::cout <<"- max_iterations: " <<max_iterations <<"\n"; 
  }

  // compute size of work space
  int G_sz = (md.G.length() - 1)/(md.m + 1);
  int max_l = 0;
  int l = 0;
  for (int i = 0; i < md.K; i++) {
    l = l + md.G_blkszs[i];
    if (max_l < md.G_blkszs[i])
      max_l = md.G_blkszs[i];
  }
  int F_sz = (md.F.length() - 1)/(md.m + 1);
  int max_n = 0;
  int n = 0;
  for (int i = 0; i < md.L; i++) {
    n = n + md.F_blkszs[i];
    if (max_n < md.F_blkszs[i])
      max_n = md.F_blkszs[i];
  }
  int sz = F_sz + G_sz;
  int M_upsz = md.m * md.m;
  int NB = 32;
  int lwork = (2*md.m+5)*sz + 2*(n+l) + 
    MAX(md.m+sz*NB,MAX(3*(md.m+M_upsz+MAX(G_sz,F_sz)),
		    MAX(3*(MAX(max_l,max_n)+MAX(G_sz,F_sz)),
			MAX(G_sz+3*max_l,F_sz+3*max_n))));

  // define G
  double* G;
  G = (double*) calloc(md.G.length() - 1, sizeof(double));
  if (G == NULL) die_rig_maxdet("G == NULL.\n");
  for (int i = 0; i < md.G.length() - 1; i++)
    G[i] = NTL::to_double(md.G[i+1])/NTL::to_double(md.G[0]);
  
  // define F
  double *F;
  F = (double*) calloc(md.F.length() - 1, sizeof(double));
  if (F == NULL) die_rig_maxdet("F == NULL.\n");
  for (int i = 0; i < md.F.length() - 1; i++)
    F[i] = NTL::to_double(md.F[i+1])/NTL::to_double(md.F[0]);
  
  // define c
  double* c;
  c = (double*) calloc(md.c.length() - 1, sizeof(double));
  if (c == NULL) die_rig_maxdet("c == NULL.\n");
  for (int i = 0; i < md.c.length() - 1; i++)
    c[i] = NTL::to_double(md.c[i+1]) / NTL::to_double(md.c[0]);
  
  // define other parameters
  double ul[2];
  double* hist;
  hist = (double*) calloc(3*max_iterations, sizeof(double));
  if (hist == NULL) die_rig_maxdet("hist == NULL.\n");
  double gamma = MAXDET_GAMMA;
  double reltol = MAXDET_RELTOL;
  double* work;
  work = (double*) calloc(lwork, sizeof(double));
  if (work == NULL) die_rig_maxdet("work == NULL.\n");
  int iwork[3];
  int info; 

  double* minimizer;
  minimizer = (double*) calloc(md.m, sizeof(double));
  if (minimizer == NULL) die_rig_maxdet("minimizer == NULL\n.");
  for (int i = 0; i < md.m; i++)
    minimizer[i] = interior_point[i];   
 
  double *dual_Z;
  dual_Z = (double*) calloc(md.F.length() - 1, sizeof(double));
  if (dual_Z == NULL) die_rig_maxdet("dual_Z == NULL\n.");
  
  double *dual_W;
  dual_W = (double*) calloc(md.G.length() - 1, sizeof(double));
  if (dual_W == NULL) die_rig_maxdet("dual_W == NULL\n.");

  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE)
    std::cout <<"- starting MAXDET (workspace = " <<lwork <<")\n";

  maxdet(md.m, md.L, F, md.F_blkszs, md.K, G, md.G_blkszs, c,
 	 minimizer, dual_Z, dual_W, ul, hist, gamma, duality_gap,
 	 reltol, const_cast<int*>(&max_iterations), work, lwork, iwork, &info,
	 output_level);

  if (info < 0) die_rig_maxdet("info < 0.\n");
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"- MAXDET stopped because ";
    switch (info) {
    case 1: std::cout <<"max_iterations exceeded (increase might help).\n"; break;
    case 2: std::cout <<"duality_gap reached.\n"; break;
    case 3: std::cout <<"reltol reached.\n"; break;
    }
  }
  if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
    std::cout <<"- primal value = " <<ul[0] <<"\n";
    std::cout <<"- dual value   = " <<ul[1] <<"\n";
    std::cout <<"- minimizer = ";
    for (int i = 0; i < md.m; i++) 
      std::cout <<minimizer[i] <<" ";
    std::cout <<"\n";
  }
  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"- dual_Z = ";
    for (int i = 0; i < F_sz; i++)
      std::cout <<dual_Z[i] <<" ";
    std::cout <<"\n";
    std::cout <<"- dual_W = ";
    for (int i = 0; i < G_sz; i++) 
      std::cout <<dual_W[i] <<" ";
    std::cout <<"\n";
  }
  
  upper_bound = ul[0];
  lower_bound = ul[1];

  free(work);
  free(hist);
  free(c);
  free(G);
  free(F);
  free(minimizer);
  free(dual_Z);
  free(dual_W);
}
