/* -*- Mode:C++; -*- */

// rig_maxdet.h

#ifndef __RIG_MAXDET_H
#define __RIG_MAXDET_H

#include "QQ.h"

#define POSLT(DIM,ROW,COL) (((ROW)*(ROW+1))/2+COL)
#define POSUT(DIM,ROW,COL) ((ROW)*(DIM)-((ROW-1)*(ROW))/2+COL-(ROW))

#define OUTPUT_LEVEL_NONE 0
#define OUTPUT_LEVEL_UNCOMMUNICATIVE 1
#define OUTPUT_LEVEL_COMMUNICATIVE 2
#define OUTPUT_LEVEL_ALL 3


// maxdet_bound is to be read as lin_part + log(det_part)

struct maxdet_bound {
  QQ det_part;
  QQ lin_part;

  bool operator>(const maxdet_bound& md_bound) const 
  {
    QQ one(1);
    if (one + lin_part - md_bound.lin_part > md_bound.det_part / det_part)
      // use the amazing fact exp(x) >= 1 + x for all x
      return true;  // this is kosher
    else
      return false; // this does not say anything
  }
};


// maxdet_data contains the data of a general maxdet problem:
//   min  c^t x - log(det G(x))
//   s.t. F(x) = F_0 + x_1 F_1 + ... + x_m F_m  pos.semi.def.
//        G(x) = G_0 + x_1 G_1 + ... + x_m G_m  pos.def.
// The matrices F_i and G_i are block diagonalized and then stored in 
// vectors F and G in a packed format (see HTML/rmd.html for an example). 

struct maxdet_data {
  int m;         // no of variables 
  int L;         // no of blocks in F 
  int *F_blkszs; // L-vector, dimensions of diagonal blocks of F 
  NTL::vec_ZZ F; // F_i's in packed storage (first entry is common denominator)
  int K;         // no of blocks in G 
  int *G_blkszs; // K-vector, dimensions of diagonal blocks of G
  NTL::vec_ZZ G; // G_i's in packed storage (first entry is common denominator)
  NTL::vec_ZZ c; // m-vector (first entry is common denominator)
};


// Given a maxdet problem in maxdet_data form, together with an interior point
// (primal feasible solution), a desired duality gap and a desired maximum
// number of iterations to be performed, the following just calls the routines
// of MAXDET. 
// (see http://www.stanford.edu/~boyd/MAXDET.html for more information)

void solve_maxdet_nonrigorously 
(
 const maxdet_data& md,
 const double* interior_point,
 const double duality_gap,
 const int max_iterations,
 double& lower_bound,
 double& upper_bound,
 const int output_level   // use constants
 );


// The input is as above, but an approximated minimizer and an approximated 
// solution (W,Z) to the dual problem is returned. The dual is of the form
// max  log (det W) - trace(G_0 W) - trace(F_0 Z) + m
// s.t. trace(G_i W) + trace(F_i Z) = c_i , i=1,..,d
//      W pos.def., Z pos.semi.def.
// So in particular, the returned Z is of the same format as the F_i and W 
// as the G_i.
// Note that the first entry in all verctors is a common denominator.

bool solve_maxdet_rigorously 
(
 const maxdet_data& md,
 const double* interior_point,
 const double duality_gap,
 const int max_iterations,
 NTL::vec_ZZ& minimizer_approx,    // first entry is common denominator
 NTL::vec_ZZ& dual_Z_approx,       // first entry is common denominator
 NTL::vec_ZZ& dual_W_approx,       // first entry is common denominator
 maxdet_bound& lower_bound,
 maxdet_bound& upper_bound,
 const int output_level            // use constants
 );


// Given an approximated minimizer, one can test primal feasibility 
// and obtain an upper_bound, if feasible. 

bool check_primal_solution
(
 const maxdet_data& md, 
 const NTL::vec_ZZ& minimizer_approx, // first entry is common denominator
 maxdet_bound& upper_bound,
 const int output_level               
 );


// Given an approximated solution of the dual (W,Z), one can test 
// primal feasibility and obtain a lower_bound, if feasible. 

bool check_dual_solution
(
 const maxdet_data& md,
 const NTL::vec_ZZ& dual_Z_approx,    // first entry is common denominator
 const NTL::vec_ZZ& dual_W_approx,    // first entry is common denominator
 maxdet_bound& lower_bound,
 const int output_level               
 );

#endif
