/* -*- Mode:C++; -*- */

// rmd.cc

#include <math.h>       
#include <iostream>
#include <NTL/mat_ZZ.h>

#include "QQ.h"
#include "rig_maxdet.h"

#define COUT_PRECISION 10

void die_rmd(char* last_words)
{
  std::cout <<last_words;
  exit(EXIT_FAILURE);
}

void get_md(maxdet_data &md)
{
  std::cin >>md.m;
  std::cin >>md.L;
  md.F_blkszs = (int*) calloc(md.L, sizeof(int));
  if (md.F_blkszs == NULL) die_rmd("md.F_blkszs == NULL.\n");
  for (int i = 0; i < md.L; i++)  
    std::cin >>md.F_blkszs[i];
  std::cin >>md.F;
  std::cin >>md.K;
  md.G_blkszs = (int*) calloc(md.K, sizeof(int));
  if (md.G_blkszs == NULL) die_rmd("md.G_blkszs == NULL.\n");
  for (int i = 0; i < md.K; i++)  
    std::cin >>md.G_blkszs[i];
  std::cin >>md.G;
  std::cin >>md.c;
}

void put_md(maxdet_data &md)
{
  std::cout <<"* md.m = " <<md.m <<"\n";
  std::cout <<"* md.L = " <<md.L <<"\n";
  std::cout <<"* md.F_blkszs = ";
  for (int i = 0; i < md.L; i++)  
    std::cout << md.F_blkszs[i] <<" ";
  std::cout <<"\n";
  std::cout <<"* md.F = " <<md.F <<"\n";
  std::cout <<"* md.K = " <<md.K <<"\n";
  std::cout <<"* md.G_blkszs = ";
  for (int i = 0; i < md.K; i++)  
    std::cout <<md.G_blkszs[i] << " ";
  std::cout <<std::endl;
  std::cout <<"* md.G = " << md.G << "\n";
  std::cout <<"* md.c = " << md.c << "\n";
}


int main(int argc, char** argv)
{
  std::cout.precision(COUT_PRECISION);

  int output_level = OUTPUT_LEVEL_COMMUNICATIVE;
  bool set_duality_gap = false; 
  bool set_maximum_iterations = false;
  bool proof_mode = false;
  bool quick_and_dirty_mode = false;
 
  int max_iterations; 
  double duality_gap; 

  int c;

  while ((c = getopt (argc, argv, "hd:pm:qvn")) != -1)
    switch (c)
      {
      case 'h':
        std::cout <<"Usage: rmd [options] <file\n";
        std::cout <<"Options:\n";
        std::cout <<" -h   show this help\n";
	std::cout <<" -d x set duality gap to x\n";
        std::cout <<" -p   proof mode\n";
	std::cout <<" -m N set maximum iterations to N\n";
        std::cout <<" -q   quick and dirty mode (non-rigorous)\n";
        std::cout <<" -v   be verbose\n";
        std::cout <<" -n   be non-verbose\n";
        return 0;
        break;   
      case 'd':
	duality_gap = atof(optarg);
	set_duality_gap = true;
	break;
      case 'p':
        proof_mode = true;
        break;
      case 'q':
        quick_and_dirty_mode = true; 
        break;   
      case 'm':
	max_iterations = atoi(optarg);
	set_maximum_iterations = true; 
	break;
      case 'n':
        output_level = OUTPUT_LEVEL_UNCOMMUNICATIVE;
        break;
      case 'v':
        output_level = OUTPUT_LEVEL_ALL;
        break;
      default:
        std::cout <<"rmd: invalid option\n";
        std::cout <<"Try 'rmd -h' for more information.\n";
        return 1;
      }

  if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
    std::cout <<"***********\n";
    std::cout <<"*** rmd ***\n";
    std::cout <<"***********\n\n";
  }
	
  // 1. read input

  if (output_level >= OUTPUT_LEVEL_ALL) {
    std::cout <<"* READING INPUT\n";
  }

  maxdet_data md; 
  get_md(md);

  if (output_level >= OUTPUT_LEVEL_ALL) {
    put_md(md);
  }

  NTL::vec_ZZ minimizer_approx;
  NTL::vec_ZZ dual_Z_approx;       
  NTL::vec_ZZ dual_W_approx;       
  maxdet_bound md_lower_bound;
  maxdet_bound md_upper_bound;
  double lower_bound;
  double upper_bound;

  double* interior;
  interior = (double*) calloc(md.m, sizeof(double));
  if (interior == NULL) die_rmd ("interior == NULL.\n");

  if (proof_mode) {
    std::cin >>minimizer_approx;
    std::cin >>dual_Z_approx;
    std::cin >>dual_W_approx;
    
    if (output_level >= OUTPUT_LEVEL_ALL) {
      std::cout <<"* minimizer_approx = " <<minimizer_approx <<"\n";
      std::cout <<"* dual_Z_approx = " <<dual_Z_approx <<"\n";
      std::cout <<"* dual_W_approx = " <<dual_W_approx <<"\n";
    }
  }
  else {
    for (int i = 0; i < md.m; i++)
      std::cin >>interior[i];
    
    if (!set_maximum_iterations)
      std::cin >>max_iterations;
    
    if (!set_duality_gap)
      std::cin >>duality_gap;

    if (output_level >= OUTPUT_LEVEL_ALL) {
      std::cout <<"* interior = ";
      for (int i = 0; i < md.m; i++)  
	std::cout << interior[i] << " ";
      std::cout << std::endl;
      std::cout <<"* max_iterations = " <<max_iterations <<"\n";
      std::cout <<"* duality_gap = " <<duality_gap <<"\n";
      std::cout <<"\n";
    }
  }

  // 2. compute in three modes: proof, quick_and_dirty, normal
  
  if (proof_mode) {
    if (!check_primal_solution(md,minimizer_approx,md_upper_bound,output_level-1))
      die_rmd("Not primal feasible.\n");
    if (!check_dual_solution(md,dual_Z_approx,dual_W_approx,md_lower_bound,output_level-1))
      die_rmd("Not dual feasible.\n");

    QQ lin_part_diff = md_upper_bound.lin_part - md_lower_bound.lin_part;
    QQ det_part_diff = md_upper_bound.det_part / md_lower_bound.det_part;

    if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
      std::cout <<"- duality_gap";      
      std::cout <<" = " <<lin_part_diff <<" + log(" <<det_part_diff <<")\n\n";
    }
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      if ((lin_part_diff == NTL::to_ZZ(0)) && (det_part_diff == NTL::to_ZZ(1))) 
        std::cout << "* CERTIFIED OPTIMUM *\n\n";
      else
	std::cout << "* NOT A CERTIFIED OPTIMUM *\n\n";
    }
  }

  else if (quick_and_dirty_mode) {      
    solve_maxdet_nonrigorously(md, interior, duality_gap, max_iterations,
			       lower_bound, upper_bound, output_level - 1);
    
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      std::cout <<"\n* lower_bound ~ " << lower_bound <<"\n";
      std::cout <<"* upper_bound ~ " << upper_bound <<"\n";
    }
  }

  else {
    solve_maxdet_rigorously(md, interior, duality_gap, max_iterations, minimizer_approx, 
			    dual_Z_approx, dual_W_approx, md_lower_bound, md_upper_bound, output_level);

    QQ lin_part_diff = md_upper_bound.lin_part - md_lower_bound.lin_part;
    QQ det_part_diff = md_upper_bound.det_part / md_lower_bound.det_part;
    
    if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
      std::cout <<"* duality_gap";
    }
    if (output_level >= OUTPUT_LEVEL_UNCOMMUNICATIVE) {
      std::cout <<" = " <<lin_part_diff <<" + log(" <<det_part_diff <<")";
    }
    if (output_level >= OUTPUT_LEVEL_COMMUNICATIVE) {
      std::cout <<" ~ " <<to_double(lin_part_diff) + log(to_double(det_part_diff)) <<"\n" ;
    }      
  }

  free(interior);    
  return 0;
}
